#include <e.h>

/* for complex link stuff */
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <netdb.h>
#include <net/ethernet.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <net/if_arp.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <net/if.h>
#include <linux/types.h>
#include <linux/wireless.h>
#include <sys/socket.h>

/***************************************************************************/
typedef struct _Instance Instance;

struct _Instance
{
   E_Gadcon_Client *gcc;
   Evas_Object *obj;
   Ecore_Poller *poller;
   int strength;
};

/***************************************************************************/
/**/
/* gadcon requirements */
static E_Gadcon_Client *_gc_init(E_Gadcon *gc, const char *name, const char *id, const char *style);
static void _gc_shutdown(E_Gadcon_Client *gcc);
static void _gc_orient(E_Gadcon_Client *gcc);
static char *_gc_label(void);
static Evas_Object *_gc_icon(Evas *evas);
static const char *_gc_id_new(void);
/* and actually define the gadcon class that this module provides (just 1) */
static const E_Gadcon_Client_Class _gadcon_class =
{
   GADCON_CLIENT_CLASS_VERSION,
     "illume-wifi",
     {
	_gc_init, _gc_shutdown, _gc_orient, _gc_label, _gc_icon, _gc_id_new, NULL
     },
   E_GADCON_CLIENT_STYLE_PLAIN
};
static E_Module *mod = NULL;
/**/
/***************************************************************************/

static int _cb_poll(void *data);
static void _cb_cleanup(void);

/* called from the module core */
void
_e_mod_gad_wifi_init(E_Module *m)
{
   mod = m;
   e_gadcon_provider_register(&_gadcon_class);
}

void
_e_mod_gad_wifi_shutdown(void)
{
   e_gadcon_provider_unregister(&_gadcon_class);
   mod = NULL;
}

/* internal calls */
static Evas_Object *
_theme_obj_new(Evas *e, const char *custom_dir, const char *group)
{
   Evas_Object *o;
   
   o = edje_object_add(e);
   if (!e_theme_edje_object_set(o, "base/theme/modules/illume", group))
     {
	if (custom_dir)
	  {
	     char buf[PATH_MAX];
	     
	     snprintf(buf, sizeof(buf), "%s/illume.edj", custom_dir);
	     if (edje_object_file_set(o, buf, group))
	       {
		  printf("OK FALLBACK %s\n", buf);
	       }
	  }
     }
   return o;
}

static E_Gadcon_Client *
_gc_init(E_Gadcon *gc, const char *name, const char *id, const char *style)
{
   Evas_Object *o;
   E_Gadcon_Client *gcc;
   Instance *inst;
   
   inst = E_NEW(Instance, 1);
   o = _theme_obj_new(gc->evas, e_module_dir_get(mod),
		      "e/modules/illume/gadget/wifi");
   evas_object_show(o);
   gcc = e_gadcon_client_new(gc, name, id, style, o);
   gcc->data = inst;
   inst->gcc = gcc;
   inst->obj = o;
   e_gadcon_client_util_menu_attach(gcc);
   
   inst->strength = -1;
   inst->poller = ecore_poller_add(ECORE_POLLER_CORE, 16, _cb_poll, inst);
   
   return gcc;
}

static void
_gc_shutdown(E_Gadcon_Client *gcc)
{
   Instance *inst;
   
   inst = gcc->data;
   _cb_cleanup();
   ecore_poller_del(inst->poller);
   evas_object_del(inst->obj);
   free(inst);
}

static void
_gc_orient(E_Gadcon_Client *gcc)
{
   Instance *inst;
   Evas_Coord mw, mh;
   
   inst = gcc->data;
   mw = 0, mh = 0;
   edje_object_size_min_get(inst->obj, &mw, &mh);
   if ((mw < 1) || (mh < 1))
     edje_object_size_min_calc(inst->obj, &mw, &mh);
   if (mw < 4) mw = 4;
   if (mh < 4) mh = 4;
   e_gadcon_client_aspect_set(gcc, mw, mh);
   e_gadcon_client_min_size_set(gcc, mw, mh);
}

static char *
_gc_label(void)
{
   return "Wifi (Illume)";
}

static Evas_Object *
_gc_icon(Evas *evas)
{
/* FIXME: need icon
   Evas_Object *o;
   char buf[4096];
   
   o = edje_object_add(evas);
   snprintf(buf, sizeof(buf), "%s/e-module-clock.edj",
	    e_module_dir_get(clock_module));
   edje_object_file_set(o, buf, "icon");
   return o;
 */
   return NULL;
}

static const char *
_gc_id_new(void)
{
   return _gadcon_class.name;
}

/* the complex way - thanks to simple being broken on gta02 */
static int
link_ext_get(int fd, const char *name, int req, struct iwreq *rq)
{
   strncpy(rq->ifr_name, name, sizeof(rq->ifr_name));
   return (ioctl(fd, req, rq));
}

static int link_fd = -1;
static link_ifname[1024];
static struct iw_range link_range;
static int link_has_range;

static int
link_setup(const char *name)
{
   struct iwreq rq;

   link_ifname[0] = 0;
   if (!name)
     {
	FILE *f;
	
	link_ifname[0] = 0;
	f = fopen("/proc/net/wireless", "r");
	if (f)
	  {
	     char buf[1024];
	     
	     while (fgets(buf, sizeof(buf), f))
	       {
		  if (strchr(buf, ':'))
		    {
		       char *p;
		       
		       sscanf(buf, "%s", link_ifname);
		       p = strchr(link_ifname, ':');
		       if (p) *p = 0;
		       break;
		    }
	       }
	     fclose(f);
	  }
     }
   else
     strcpy(link_ifname, name);

   printf("ifname: %s\n", link_ifname);
   if (!link_ifname[0]) return -1;
   
   link_fd = socket(AF_INET, SOCK_DGRAM, 0);
   if (link_fd < 0) return -1;
   printf("link_fd: %i\n", link_fd);
   
   link_has_range = 0;
   rq.u.data.pointer = (caddr_t)(&link_range);
   rq.u.data.length = sizeof(link_range);
   rq.u.data.flags = 1;
   
   if (link_ext_get(link_fd, link_ifname, SIOCGIWRANGE, &rq) >= 0)
     {
	printf("has range ok\n");
	link_has_range = 1;
	return 0;
     }
   close(link_fd);
   link_fd = -1;
   return -1;
}

static double
link_quality(void)
{
   struct iwreq rq;
   struct iw_statistics stats;
   
   if (link_fd < 0) return -1;
   
   printf("get link qual...\n");
   rq.u.data.pointer = (caddr_t)(&stats);
   rq.u.data.length = sizeof(stats);
   rq.u.data.flags = 1;
   if (link_ext_get(link_fd, link_ifname, SIOCGIWSTATS, &rq) >= 0)
     {
	printf("... %i\n", stats.qual.qual);
	return (double)stats.qual.qual / (double)link_range.max_qual.qual;
     }
   printf("...FAIL\n");
   close(link_fd);
   link_fd = -1;
   return -1;
}

static void
link_end(void)
{
   if (link_fd < 0) return;
   printf("link end\n");
   close(link_fd);
   link_fd = -1;
}

/* the simple way */
static int
_wifi_signal_get(void)
{
   FILE *f;
   int strength = -1;
   char buf[4096];
   
   f = fopen("/proc/net/wireless", "r");
   if (!f) return -1;
   
   while (fgets(buf, 256, f))
     {
	int dummy;
	int count;
	int found_dev = 0;
	int wlan_status = 0;
	int wlan_link = 0;
	int wlan_level = 0;
	int wlan_noise = 0;
	char iface[64];
	
	count = sscanf(buf, "%s %i %i %i %i %i %i %i %i %i %i\n",
		       iface, &wlan_status, &wlan_link, &wlan_level, 
		       &wlan_noise, &dummy, &dummy, &dummy, &dummy, &dummy, 
		       &dummy);
	if (count != 3) continue;
	fclose(f);
	return wlan_link;
     }
   fclose(f);
   return -1;
}

static int
_cb_poll(void *data)
{
   Instance *inst;
   int pstrength;
   
   inst = data;
   pstrength = inst->strength;
/* complex way */
   if (link_fd < 0)
     {
	if (link_setup(NULL) >= 0)
	  inst->strength = link_quality() * 100.0;
     }
   else
     inst->strength = link_quality() * 100.0;
   
/* simple way - not working */
/*   inst->strength = _wifi_signal_get(); */
   if (inst->strength != pstrength)
     {
	Edje_Message_Float msg;
	double level;
	
	level = (double)inst->strength / 100.0;
	if (level < 0.0) level = 0.0;
	else if (level > 1.0) level = 1.0;
	msg.val = level;
	edje_object_message_send(inst->obj, EDJE_MESSAGE_FLOAT, 1, &msg);
     }
   return 1;
}

static void
_cb_cleanup(void)
{
   link_end();
}
